# -*- coding: utf-8 -*-
from girder import events
from girder.plugin import GirderPlugin

from .quota import QuotaPolicy


class UserQuotaPlugin(GirderPlugin):
    DISPLAY_NAME = 'User and Collection Quotas'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):
        quota = QuotaPolicy()
        info['apiRoot'].collection.route('GET', (':id', 'quota'), quota.getCollectionQuota)
        info['apiRoot'].collection.route('PUT', (':id', 'quota'), quota.setCollectionQuota)
        info['apiRoot'].user.route('GET', (':id', 'quota'), quota.getUserQuota)
        info['apiRoot'].user.route('PUT', (':id', 'quota'), quota.setUserQuota)

        events.bind('model.upload.assetstore', 'userQuota', quota.getUploadAssetstore)
        events.bind('model.upload.save', 'userQuota', quota.checkUploadStart)
        events.bind('model.upload.finalize', 'userQuota', quota.checkUploadFinalize)



from girder.exceptions import ValidationException
from girder.utility import setting_utilities


class PluginSettings:
    LICENSES = 'item_licenses.licenses'


@setting_utilities.default(PluginSettings.LICENSES)
def _defaultLicenses():
    return [
        {
            # Selected open source licenses from:
            # - https://github.com/ufal/public-license-selector/tree/57e31db
            'category': 'Code Licenses',
            'licenses': [
                {
                    'name': 'Affero General Public License 3 (AGPL-3.0)'
                },
                {
                    'name': 'Apache License 2'
                },
                {
                    'name': 'The BSD 2-Clause "Simplified" or "FreeBSD" '
                            'License'
                },
                {
                    'name': 'The BSD 3-Clause "New" or "Revised" License '
                            '(BSD)'
                },
                {
                    'name': 'Common Development and Distribution License '
                            '(CDDL-1.0)'
                },
                {
                    'name': 'Eclipse Public License 1.0 (EPL-1.0)'
                },
                {
                    'name': 'GNU General Public License 2 or later '
                            '(GPL-2.0)'
                },
                {
                    'name': 'GNU General Public License 3 (GPL-3.0)'
                },
                {
                    'name': 'GNU Library or "Lesser" General Public '
                            'License 2.1 or later (LGPL-2.1)'
                },
                {
                    'name': 'GNU Library or "Lesser" General Public '
                            'License 3.0 (LGPL-3.0)'
                },
                {
                    'name': 'The MIT License (MIT)'
                },
                {
                    'name': 'Mozilla Public License 2.0'
                }
            ]
        },
        {
            # Licenses from:
            # - http://creativecommons.org/licenses/
            #
            # Names match those from:
            # https://github.com/ufal/public-license-selector/tree/57e31db
            'category': 'Content Licenses',
            'licenses': [
                {
                    'name': 'Public Domain Dedication (CC Zero)'
                },
                {
                    'name': 'Creative Commons Attribution (CC-BY)'
                },
                {
                    'name': 'Creative Commons Attribution-ShareAlike '
                            '(CC-BY-SA)'
                },
                {
                    'name': 'Creative Commons Attribution-NoDerivs '
                            '(CC-BY-ND)'
                },
                {
                    'name': 'Creative Commons Attribution-NonCommercial '
                            '(CC-BY-NC)'
                },
                {
                    'name': 'Creative Commons Attribution-NonCommercial-'
                            'ShareAlike (CC-BY-NC-SA)'
                },
                {
                    'name': 'Creative Commons Attribution-NonCommercial-'
                            'NoDerivs (CC-BY-NC-ND)'
                },
                {
                    'name': 'Public Domain Mark (PD)'
                },
                {
                    'name': 'All Rights Reserved'
                }
            ]
        }
    ]


@setting_utilities.validator(PluginSettings.LICENSES)
def _validateLicenses(doc):
    val = doc['value']
    if not isinstance(val, list):
        raise ValidationException('Licenses setting must be a list.', 'value')
    for item in val:
        category = item.get('category', None)
        if not category or not isinstance(category, str):
            raise ValidationException(
                'License category is required and must be a non-empty string.', 'category')
        licenses = item.get('licenses', None)
        if not isinstance(licenses, list):
            raise ValidationException('Licenses in category must be a list.', 'licenses')
        for license in licenses:
            if not isinstance(license, dict):
                raise ValidationException('License must be a dict.', 'license')
            name = license.get('name', None)
            if not name or not isinstance(name, str):
                raise ValidationException(
                    'License name is required and must be a non-empty string.', 'name')



# -*- coding: utf-8 -*-
# Utility for saving and restoring the internal state of a hash object
# so that checksums can be streamed without having to remain in memory.
# Inspired by http://code.activestate.com/recipes/
# 578479-save-and-restore-sha-512-internal-state/

import hashlib
import ctypes
import binascii
import sys

_ver = sys.version_info
_HASHLIB_INLINE_EVP_STRUCT = _ver < (2, 7, 13) or (_ver >= (3,) and _ver < (3, 5, 3))
# This is the offset to the EVP_MD_CTX structure in CPython's _hashopenssl.c
# EVPObject.  It changed for Python 3.8 in
# https://github.com/python/cpython/pull/16023
_HASHLIB_EVP_STRUCT_OFFSET = 3 if _ver < (3, 8) else 2


def _getHashStateDataPointer(hashObject):
    """
    Get a pointer to the internal hash state of a hash object.

    :param hashObject: A hashlib hash object.
    :return: A ctypes pointer to the internal hash state.
    :rtype: ctypes.POINTER(ctypes.c_char)
    """
    hashPointer = ctypes.cast(id(hashObject), ctypes.POINTER(ctypes.c_void_p))

    if _HASHLIB_INLINE_EVP_STRUCT:
        # From
        #   * github.com/python/cpython/blob/2.7/Modules/_hashopenssl.c#L58
        # the layout of a hashObject (_hashlib.HASH) is:
        #     typedef struct {
        #         PyObject_HEAD
        #         PyObject *name;
        #         EVP_MD_CTX ctx;
        #         ...
        #     } EVPobject;
        #
        # Using
        #   * docs.python.org/2/c-api/structures.html#c.PyObject_HEAD
        #   * github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/evp/evp.h#L265
        # this expands to:
        #     typedef struct {
        #         Py_ssize_t ob_refcnt;      // Word 0
        #         PyTypeObject *ob_type;     // Word 1
        #         PyObject *name;            // Word 2
        #         struct env_md_ctx_st {
        #             const EVP_MD *digest;  // Word 3
        #             ENGINE *engine;        // Word 4
        #             unsigned long flags;   // Word 5
        #             void *md_data;         // Word 6   << this guy is what we want
        #             ...
        #         } /* EVP_MD_CTX */;
        #         ...
        #     } EVPobject;
        #
        # Thus, an offset of 6 words ( sizeof(void*) is 1 word) is required
        stateDataPointer = ctypes.cast(hashPointer[6], ctypes.POINTER(ctypes.c_char))
    else:
        # In cpython 2.7.13, hashlib changed to store a pointer to the OpenSSL hash
        # object rather than inlining it in the struct, so we require an extra dereference. See
        # https://github.com/python/cpython/commit/9d9615f6782be4b1f38b47d4d56cee208c26a970
        evpStruct = ctypes.cast(hashPointer[_HASHLIB_EVP_STRUCT_OFFSET],
                                ctypes.POINTER(ctypes.c_void_p))
        stateDataPointer = ctypes.cast(evpStruct[3], ctypes.POINTER(ctypes.c_char))

    assert stateDataPointer
    return stateDataPointer


_HASH_INFOS = dict()


class _HashInfo:
    def __init__(self, type, stateSize, initVectorFirstWord):
        self.type = type
        self.stateSize = stateSize

        # Instantiate to get the name and do a sanity check
        instance = self.type()
        self.name = instance.name

        stateDataPointer = _getHashStateDataPointer(instance)
        assert stateDataPointer[:8] == initVectorFirstWord

        _HASH_INFOS[self.name] = self


_HashInfo(
    type=hashlib.md5,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/md5/md5.h#L100
    stateSize=92,
    # Found empirically
    initVectorFirstWord=b'\x01\x23\x45\x67\x89\xab\xcd\xef'
)
_HashInfo(
    type=hashlib.sha1,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha.h#L100
    stateSize=96,
    # Found empirically
    initVectorFirstWord=b'\x01\x23\x45\x67\x89\xab\xcd\xef'
)
_HashInfo(
    type=hashlib.sha224,
    # Uses the same struct as sha256
    stateSize=112,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha256.c#L22
    initVectorFirstWord=b'\xd8\x9e\x05\xc1\x07\xd5\x7c\x36'
)
_HashInfo(
    type=hashlib.sha256,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha.h#L135
    stateSize=112,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha256.c#L33
    initVectorFirstWord=b'\x67\xe6\x09\x6a\x85\xae\x67\xbb'
)
_HashInfo(
    type=hashlib.sha384,
    # Uses the same struct as sha512
    stateSize=216,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha512.c#L64
    initVectorFirstWord=b'\xd8\x9e\x05\xc1\x5d\x9d\xbb\xcb'
)
_HashInfo(
    type=hashlib.sha512,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha.h#L182
    stateSize=216,
    # github.com/openssl/openssl/blob/OpenSSL_1_0_1f/crypto/sha/sha512.c#L80
    initVectorFirstWord=b'\x08\xc9\xbc\xf3\x67\xe6\x09\x6a'
)


def serialize(hashObject):
    """
    Serializes the internal state of the hash object passed in. Calling
    restore() on the serialized value afterward will return an equivalent
    hash object to the one passed.

    :param hashObject: The hash object to serialize.
    :type hashObject: _hashlib.HASH
    :returns: A binary string representing the serialized state.
    """
    stateDataPointer = _getHashStateDataPointer(hashObject)
    hashInfo = _HASH_INFOS[hashObject.name]
    return stateDataPointer[:hashInfo.stateSize]


def restore(oldStateData, hashName):
    """
    Unserializes a hash object from a binary string created by the
    serialize() method in this module.

    :param oldStateData: The serialized state.
    :type oldStateData: binary string
    :param hashName: The name of the hash algorithm type.
    :type hashName: str
    :returns: A hash object in the same state as the one that serialize()
              was called on.
    """
    hashInfo = _HASH_INFOS[hashName]
    assert len(oldStateData) == hashInfo.stateSize

    hashObject = hashInfo.type()
    stateDataPointer = _getHashStateDataPointer(hashObject)

    for i, byte in enumerate(oldStateData):
        stateDataPointer[i] = byte

    return hashObject


def serializeHex(hashObject):
    return binascii.b2a_hex(serialize(hashObject))


def restoreHex(oldHexStateData, hashName):
    return restore(binascii.a2b_hex(oldHexStateData), hashName)



from girder.exceptions import ValidationException
from girder.utility import setting_utilities


class PluginSettings:
    TRACKING_ID = 'google_analytics.tracking_id'


@setting_utilities.default(PluginSettings.TRACKING_ID)
def _defaultTrackingId():
    return ''


@setting_utilities.validator(PluginSettings.TRACKING_ID)
def _validateTrackingId(doc):
    if not doc['value']:
        raise ValidationException('Google Analytics Tracking ID must not be empty.', 'value')



# -*- coding: utf-8 -*-
from girder import events
from girder.constants import AccessType
from girder.models.file import File
from girder.plugin import GirderPlugin


def _onDownloadFileRequest(event):
    if event.info['startByte'] == 0:
        File().increment(
            query={'_id': event.info['file']['_id']},
            field='downloadStatistics.started',
            amount=1)
    File().increment(
        query={'_id': event.info['file']['_id']},
        field='downloadStatistics.requested',
        amount=1)


def _onDownloadFileComplete(event):
    File().increment(
        query={'_id': event.info['file']['_id']},
        field='downloadStatistics.completed',
        amount=1)


class DownloadStatisticsPlugin(GirderPlugin):
    DISPLAY_NAME = 'Download Statistics'

    def load(self, info):
        # Bind REST events
        events.bind('model.file.download.request', 'download_statistics', _onDownloadFileRequest)
        events.bind('model.file.download.complete', 'download_statistics', _onDownloadFileComplete)

        # Add download count fields to file model
        File().exposeFields(level=AccessType.READ, fields='downloadStatistics')



# -*- coding: utf-8 -*-
from collections import OrderedDict

import cherrypy
from bson import ObjectId
import re

from girder.constants import GIRDER_ROUTE_ID
from girder.exceptions import ValidationException
from girder.utility import setting_utilities


class SettingKey:
    """
    Core settings should be enumerated here by a set of constants corresponding
    to sensible strings.
    """

    ADD_TO_GROUP_POLICY = 'core.add_to_group_policy'
    API_KEYS = 'core.api_keys'
    BANNER_COLOR = 'core.banner_color'
    BRAND_NAME = 'core.brand_name'
    COLLECTION_CREATE_POLICY = 'core.collection_create_policy'
    CONTACT_EMAIL_ADDRESS = 'core.contact_email_address'
    COOKIE_LIFETIME = 'core.cookie_lifetime'
    CORS_ALLOW_HEADERS = 'core.cors.allow_headers'
    CORS_ALLOW_METHODS = 'core.cors.allow_methods'
    CORS_ALLOW_ORIGIN = 'core.cors.allow_origin'
    CORS_EXPOSE_HEADERS = 'core.cors.expose_headers'
    EMAIL_FROM_ADDRESS = 'core.email_from_address'
    EMAIL_HOST = 'core.email_host'
    EMAIL_VERIFICATION = 'core.email_verification'
    ENABLE_NOTIFICATION_STREAM = 'core.enable_notification_stream'
    ENABLE_PASSWORD_LOGIN = 'core.enable_password_login'
    GIRDER_MOUNT_INFORMATION = 'core.girder_mount_information'
    PRIVACY_NOTICE = 'core.privacy_notice'
    REGISTRATION_POLICY = 'core.registration_policy'
    ROUTE_TABLE = 'core.route_table'
    SERVER_ROOT = 'core.server_root'
    SMTP_ENCRYPTION = 'core.smtp.encryption'
    SMTP_HOST = 'core.smtp_host'
    SMTP_PASSWORD = 'core.smtp.password'
    SMTP_PORT = 'core.smtp.port'
    SMTP_USERNAME = 'core.smtp.username'
    UPLOAD_MINIMUM_CHUNK_SIZE = 'core.upload_minimum_chunk_size'
    USER_DEFAULT_FOLDERS = 'core.user_default_folders'


class SettingDefault:
    """
    Core settings that have a default should be enumerated here with the
    SettingKey.
    """

    defaults = {
        SettingKey.ADD_TO_GROUP_POLICY: 'never',
        SettingKey.API_KEYS: True,
        SettingKey.BANNER_COLOR: '#3F3B3B',
        SettingKey.BRAND_NAME: 'Girder',
        SettingKey.COLLECTION_CREATE_POLICY: {
            'open': False,
            'groups': [],
            'users': []
        },
        SettingKey.CONTACT_EMAIL_ADDRESS: 'kitware@kitware.com',
        SettingKey.COOKIE_LIFETIME: 180,
        # These headers are necessary to allow the web server to work with just
        # changes to the CORS origin
        SettingKey.CORS_ALLOW_HEADERS:
            'Accept-Encoding, Authorization, Content-Disposition, '
            'Content-Type, Cookie, Girder-Authorization, Girder-OTP, Girder-Token',
        SettingKey.CORS_ALLOW_METHODS: 'GET, POST, PUT, HEAD, DELETE',
        SettingKey.CORS_ALLOW_ORIGIN: '',
        SettingKey.CORS_EXPOSE_HEADERS: 'Girder-Total-Count',
        # An apache server using reverse proxy would also need
        #  X-Requested-With, X-Forwarded-Server, X-Forwarded-For,
        #  X-Forwarded-Host, Remote-Addr
        SettingKey.EMAIL_FROM_ADDRESS: 'Girder <no-reply@girder.org>',
        # SettingKey.EMAIL_HOST is provided by a function
        SettingKey.EMAIL_VERIFICATION: 'disabled',
        SettingKey.ENABLE_NOTIFICATION_STREAM: True,
        SettingKey.ENABLE_PASSWORD_LOGIN: True,
        SettingKey.GIRDER_MOUNT_INFORMATION: None,
        SettingKey.PRIVACY_NOTICE: 'https://www.kitware.com/privacy',
        SettingKey.REGISTRATION_POLICY: 'open',
        # SettingKey.ROUTE_TABLE is provided by a function
        SettingKey.SERVER_ROOT: '',
        SettingKey.SMTP_ENCRYPTION: 'none',
        SettingKey.SMTP_HOST: 'localhost',
        SettingKey.SMTP_PASSWORD: '',
        SettingKey.SMTP_PORT: 25,
        SettingKey.SMTP_USERNAME: '',
        SettingKey.UPLOAD_MINIMUM_CHUNK_SIZE: 1024 * 1024 * 5,
        SettingKey.USER_DEFAULT_FOLDERS: 'public_private'
    }

    @staticmethod
    @setting_utilities.default(SettingKey.EMAIL_HOST)
    def _defaultEmailHost():
        if cherrypy.request and cherrypy.request.local and cherrypy.request.local.name:
            host = '%s://%s' % (cherrypy.request.scheme, cherrypy.request.local.name)
            if cherrypy.request.local.port != 80:
                host += ':%d' % cherrypy.request.local.port
            return host

    @staticmethod
    @setting_utilities.default(SettingKey.ROUTE_TABLE)
    def _defaultRouteTable():
        return {
            GIRDER_ROUTE_ID: '/'
        }


class SettingValidator:
    @staticmethod
    @setting_utilities.validator(SettingKey.ADD_TO_GROUP_POLICY)
    def _validateAddToGroupPolicy(doc):
        doc['value'] = doc['value'].lower()
        if doc['value'] not in ('never', 'noadmin', 'nomod', 'yesadmin', 'yesmod', ''):
            raise ValidationException(
                'Add to group policy must be one of "never", "noadmin", '
                '"nomod", "yesadmin", or "yesmod".', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.API_KEYS)
    def _validateApiKeys(doc):
        if not isinstance(doc['value'], bool):
            raise ValidationException('API key setting must be boolean.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.BANNER_COLOR)
    def _validateBannerColor(doc):
        if not doc['value']:
            raise ValidationException('The banner color may not be empty', 'value')
        elif not (re.match(r'^#[0-9A-Fa-f]{6}$', doc['value'])):
            raise ValidationException('The banner color must be a hex color triplet', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.BRAND_NAME)
    def _validateBrandName(doc):
        if not doc['value']:
            raise ValidationException('The brand name may not be empty', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.COLLECTION_CREATE_POLICY)
    def _validateCollectionCreatePolicy(doc):
        from girder.models.group import Group
        from girder.models.user import User

        value = doc['value']

        if not isinstance(value, dict):
            raise ValidationException('Collection creation policy must be a JSON object.', 'value')

        for i, groupId in enumerate(value.get('groups', ())):
            Group().load(groupId, force=True, exc=True)
            value['groups'][i] = ObjectId(value['groups'][i])

        for i, userId in enumerate(value.get('users', ())):
            User().load(userId, force=True, exc=True)
            value['users'][i] = ObjectId(value['users'][i])

        value['open'] = value.get('open', False)

    @staticmethod
    @setting_utilities.validator(SettingKey.CONTACT_EMAIL_ADDRESS)
    def _validateContactEmailAddress(doc):
        # This is typically used within an RFC 6068 "mailto:" scheme, so no display name is allowed
        from girder.utility.mail_utils import validateEmailAddress
        if not validateEmailAddress(doc['value']):
            raise ValidationException(
                'Contact email address must be a valid email address.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.COOKIE_LIFETIME)
    def _validateCookieLifetime(doc):
        try:
            doc['value'] = int(doc['value'])
            if doc['value'] > 0:
                return
        except ValueError:
            pass  # We want to raise the ValidationException
        raise ValidationException('Cookie lifetime must be an integer > 0.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.CORS_ALLOW_HEADERS)
    def _validateCorsAllowHeaders(doc):
        if isinstance(doc['value'], str):
            headers = doc['value'].replace(',', ' ').strip().split()
            # remove duplicates
            headers = list(OrderedDict.fromkeys(headers))
            doc['value'] = ', '.join(headers)
            return
        raise ValidationException(
            'Allowed headers must be a comma-separated list or an empty string.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.CORS_ALLOW_METHODS)
    def _validateCorsAllowMethods(doc):
        if isinstance(doc['value'], str):
            methods = doc['value'].replace(',', ' ').strip().upper().split()
            # remove duplicates
            methods = list(OrderedDict.fromkeys(methods))
            doc['value'] = ', '.join(methods)
            return
        raise ValidationException(
            'Allowed methods must be a comma-separated list or an empty string.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.CORS_ALLOW_ORIGIN)
    def _validateCorsAllowOrigin(doc):
        if isinstance(doc['value'], str):
            origins = doc['value'].replace(',', ' ').strip().split()
            origins = [origin.rstrip('/') for origin in origins]
            # remove duplicates
            origins = list(OrderedDict.fromkeys(origins))
            doc['value'] = ', '.join(origins)
            return
        raise ValidationException(
            'Allowed origin must be a comma-separated list of base urls or * or an empty string.',
            'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.CORS_EXPOSE_HEADERS)
    def _validateCorsExposeHeaders(doc):
        if not isinstance(doc['value'], str):
            raise ValidationException('CORS exposed headers must be a string', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.EMAIL_FROM_ADDRESS)
    def _validateEmailFromAddress(doc):
        # mail_utils.validateEmailAddress cannot be used here, as RFC 5322 allows this to accept an
        # an address which includes a display name too
        if not doc['value']:
            raise ValidationException('Email from address must not be blank.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.EMAIL_HOST)
    def _validateEmailHost(doc):
        if isinstance(doc['value'], str):
            doc['value'] = doc['value'].strip()
            return
        raise ValidationException('Email host must be a string.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.EMAIL_VERIFICATION)
    def _validateEmailVerification(doc):
        doc['value'] = doc['value'].lower()
        if doc['value'] not in ('required', 'optional', 'disabled'):
            raise ValidationException(
                'Email verification must be "required", "optional", or "disabled".', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.ENABLE_NOTIFICATION_STREAM)
    def _validateEnableNotificationStream(doc):
        if not isinstance(doc['value'], bool):
            raise ValidationException('Enable notification stream option must be boolean.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.ENABLE_PASSWORD_LOGIN)
    def _validateEnablePasswordLogin(doc):
        if not isinstance(doc['value'], bool):
            raise ValidationException('Enable password login setting must be boolean.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.GIRDER_MOUNT_INFORMATION)
    def _validateGirderMountInformation(doc):
        value = doc['value']
        if not isinstance(value, dict) or 'path' not in value:
            raise ValidationException(
                'Girder mount information must be a dict with the "path" key.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.PRIVACY_NOTICE)
    def _validatePrivacyNotice(doc):
        if not doc['value']:
            raise ValidationException('The privacy notice link may not be empty', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.REGISTRATION_POLICY)
    def _validateRegistrationPolicy(doc):
        doc['value'] = doc['value'].lower()
        if doc['value'] not in ('open', 'closed', 'approve'):
            raise ValidationException(
                'Registration policy must be "open", "closed", or "approve".', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.ROUTE_TABLE)
    def _validateRouteTable(doc):
        nonEmptyRoutes = [route for route in doc['value'].values() if route]
        if GIRDER_ROUTE_ID not in doc['value'] or not doc['value'][GIRDER_ROUTE_ID]:
            raise ValidationException('Girder root must be routable.', 'value')

        for key in doc['value']:
            if (doc['value'][key] and not doc['value'][key].startswith('/')):
                raise ValidationException('Routes must begin with a forward slash.', 'value')

        if len(nonEmptyRoutes) > len(set(nonEmptyRoutes)):
            raise ValidationException('Routes must be unique.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SERVER_ROOT)
    def _validateServerRoot(doc):
        val = doc['value']
        if val and not val.startswith('http://') and not val.startswith('https://'):
            raise ValidationException('Server root must start with http:// or https://.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SMTP_ENCRYPTION)
    def _validateSmtpEncryption(doc):
        if not doc['value'] in ['none', 'starttls', 'ssl']:
            raise ValidationException(
                'SMTP encryption must be one of "none", "starttls", or "ssl".', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SMTP_HOST)
    def _validateSmtpHost(doc):
        if not doc['value']:
            raise ValidationException('SMTP host must not be blank.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SMTP_PASSWORD)
    def _validateSmtpPassword(doc):
        if not isinstance(doc['value'], str):
            raise ValidationException('SMTP password must be a string', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SMTP_PORT)
    def _validateSmtpPort(doc):
        try:
            doc['value'] = int(doc['value'])
            if doc['value'] > 0:
                return
        except ValueError:
            pass  # We want to raise the ValidationException
        raise ValidationException('SMTP port must be an integer > 0.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.SMTP_USERNAME)
    def _validateSmtpUsername(doc):
        if not isinstance(doc['value'], str):
            raise ValidationException('SMTP username must be a string', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.UPLOAD_MINIMUM_CHUNK_SIZE)
    def _validateUploadMinimumChunkSize(doc):
        try:
            doc['value'] = int(doc['value'])
            if doc['value'] >= 0:
                return
        except ValueError:
            pass  # We want to raise the ValidationException
        raise ValidationException('Upload minimum chunk size must be an integer >= 0.', 'value')

    @staticmethod
    @setting_utilities.validator(SettingKey.USER_DEFAULT_FOLDERS)
    def _validateUserDefaultFolders(doc):
        if doc['value'] not in ('public_private', 'none'):
            raise ValidationException(
                'User default folders must be either "public_private" or "none".', 'value')
